@@ -1,6 +1,6 @@
 ;;; ebdb-mua.el --- Mail user agent interaction for EBDB  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2016-2018  Free Software Foundation, Inc.
+;; Copyright (C) 2016-2020  Free Software Foundation, Inc.
 
 ;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>
 
@@ -90,16 +90,45 @@ It can take one of the following values:
 
 Note that this option only controls how EBDB acts on information
 in incoming messages; the option `ebdb-mua-pop-up' controls
-whether the records in question are actually displayed or not."
+whether the records in question are actually displayed or not.
+
+Also see `ebdb-mua-reader-update-p' and
+`ebdb-mua-sender-update-p', as well as the equivalent values for
+each individual MUA package."
 
   ;; Also: Used for communication between `ebdb-update-records'
   ;; and `ebdb-query-create'.
-  :group 'ebdb-mua
   :type '(choice (const :tag "do nothing" nil)
-                 (const :tag "search for existing records" search)
+                 (const :tag "search for existing records" existing)
                  (const :tag "update existing records" update)
-                 (const :tag "query annotation of all messages" query)
-                 (const :tag "annotate all messages" create)
+                 (const :tag "query for update or record creation" query)
+                 (const :tag "update or create automatically" create)
+                 (function :tag "User-defined function")))
+
+(defcustom ebdb-mua-reader-update-p ebdb-mua-auto-update-p
+  "Value of `ebdb-mua-auto-update-p' for reader MUAs.
+\"Reader\" MUAs are those that provide an interface for reading
+incoming messages and articles.
+
+Defaults to the value of `ebdb-mua-auto-update-p'."
+  :type '(choice (const :tag "do nothing" nil)
+                 (const :tag "search for existing records" existing)
+                 (const :tag "update existing records" update)
+                 (const :tag "query for update or record creation" query)
+                 (const :tag "update or create automatically" create)
+                 (function :tag "User-defined function")))
+
+(defcustom ebdb-mua-sender-update-p ebdb-mua-auto-update-p
+  "Value of `ebdb-mua-auto-update-p' for sender MUAs.
+\"Sender\" MUAs are those that govern mail composition.  EBDB
+currently only supports the \"message\" and \"mail\" MUAs.
+
+Defaults to the value of `ebdb-mua-auto-update-p'."
+  :type '(choice (const :tag "do nothing" nil)
+                 (const :tag "search for existing records" existing)
+                 (const :tag "update existing records" update)
+                 (const :tag "query for update or record creation" query)
+                 (const :tag "update or create automatically" create)
                  (function :tag "User-defined function")))
 
 (defcustom ebdb-message-headers
@@ -216,6 +245,50 @@ will not operate on your own record.  See also
 		 (const :tag "Use the value of `message-alternative-emails'" message)
 		 (regexp :tag "Regexp matching your mail addresses")))
 
+(defcustom ebdb-permanent-ignores-file
+  (locate-user-emacs-file ".ebdb-permanent-ignores")
+  "File in which to save a list of permanently-ignored mails.
+EBDB can offer to permanently ignore a mail address, so that it
+will never again be considered for record creation or update.
+This option specifies the file in which to save those mails, or
+nil not to permanently ignore any mail addresses.
+
+Ignored mails are written one per line, with no name or other
+leading/trailing characters."
+  :type '(choice (const :tag "do not save ignored mails" nil)
+		 (file :tag "file in which to save ignored mails")))
+
+;; Maybe more efficient to make this a buffer instead of a list
+;; variable, and use `search-forward' to find mails instead of
+;; `member'?
+(defvar ebdb-permanently-ignored-mails nil
+  "Variable holding a list of permanently-ignored mails.")
+
+(defun ebdb-mua-load-permanent-ignores ()
+  "Load permanent ignores.
+Reads mail addresses to permanently ignore from the option
+`ebdb-permanent-ignores-file', and stores them in the variable
+`ebdb-permanently-ignored-mails'."
+  (when (and ebdb-permanent-ignores-file
+	     (file-exists-p ebdb-permanent-ignores-file))
+    (with-temp-buffer
+      (insert-file-contents ebdb-permanent-ignores-file)
+      (when (null (zerop (buffer-size)))
+	(setq ebdb-permanently-ignored-mails
+	      (split-string (buffer-string) "\n" t "[[:blank:]]"))))))
+
+(add-hook 'ebdb-after-load-hook #'ebdb-mua-load-permanent-ignores)
+
+(defun ebdb-mua-save-permanent-ignores ()
+  "Write the list of permanently-ignored mails to disk."
+  (when (and ebdb-permanent-ignores-file
+	     ebdb-permanently-ignored-mails)
+    (with-temp-file ebdb-permanent-ignores-file
+      (dolist (m ebdb-permanently-ignored-mails)
+	(insert m "\n")))))
+
+(add-hook 'ebdb-after-save-hook #'ebdb-mua-save-permanent-ignores)
+
 ;; This is currently only called in `ebdb-mua-test-headers'.
 (defun ebdb-get-user-mail-address-re ()
   "Get or set the value of variable `ebdb-user-mail-address-re'.
@@ -229,9 +302,7 @@ accordingly."
 	   (unless self-rec
 	     (user-error "`ebdb-user-mail-address-re' set to 'self, but `ebdb-record-self' is not set"))
 	   (setq ebdb-user-mail-address-re
-		 (regexp-opt (slot-value
-			      (ebdb-record-cache self-rec)
-			      'mail-canon)))))
+		 (regexp-opt (ebdb-record-mail-canon self-rec)))))
 	((and (eq ebdb-user-mail-address-re 'message)
 	      (boundp 'message-alternative-emails))
 	 (setq ebdb-user-mail-address-re
@@ -403,9 +474,10 @@ are not \"noticed\", nor is the timestamp updated."
 
 (defcustom ebdb-mua-default-formatter ebdb-default-multiline-formatter
   "The default formatter to use for MUA pop-up buffers.
-The value should be an instance of the `ebdb-formatter-ebdb'
-class.  Easy choices are the value of
-`ebdb-default-multiline-formatter' or
+The value should be an instance of the
+`ebdb-formatter-ebdb-multiline' or the
+`ebdb-formatter-ebdb-oneline' classes.  Easy choices are the
+value of `ebdb-default-multiline-formatter' or
 `ebdb-default-oneline-formatter'."
   :group 'ebdb-mua
   :type 'ebdb-formatter-ebdb)
@@ -533,7 +605,7 @@ variable should be set before EBDB is loaded.")
   (unless (plist-get slots :folder)
     (setq slots (plist-put slots :folder
 			   (ebdb-read-string
-			    "Folder name: "
+			    "Folder name"
 			    (when obj (slot-value obj 'folder))
 			    ebdb-mail-folder-list))))
   (cl-call-next-method c slots obj))
@@ -649,7 +721,8 @@ are discarded as appropriate."
   (let ((message-headers (if header-class
                              (list (assoc header-class ebdb-message-headers))
                            ebdb-message-headers))
-        address-list mail mail-list content)
+	(mail-list (copy-sequence ebdb-permanently-ignored-mails))
+        address-list mail content)
     (condition-case nil
 	(dolist (headers message-headers)
 	  (dolist (header (cdr headers))
@@ -659,14 +732,19 @@ are discarded as appropriate."
 		(setq mail (cadr address))
 		;; Ignore addresses that should be ignored.
 		(when (and mail
-			   (not (member-ignore-case mail mail-list))
+			   (not (member (downcase mail) mail-list))
+			   (not (member (downcase
+					 (concat
+					  "@"
+					  (nth 1 (split-string mail "@"))))
+					mail-list))
 			   (ebdb-mua-test-headers (car headers) address ignore-address))
 		  ;; Add each address only once. (Use MAIL-LIST for book keeping.)
 		  ;; Thus if we care about whether an address gets associated with
 		  ;; one or another header, the order of elements in
 		  ;; `ebdb-message-headers' is relevant.  The "most important"
 		  ;; headers should be first in `ebdb-message-headers'.
-		  (push mail mail-list)
+		  (push (downcase mail) mail-list)
 		  (push (list (car address) (cadr address) header (car headers) major-mode) address-list))))))
       (cl-no-applicable-method
        ;; Potentially triggered by `ebdb-mua-message-header', which
@@ -741,6 +819,23 @@ Usually this function is called by the wrapper `ebdb-mua-auto-update'."
 		  nil)))
 	  (cond ((eq task 'quit)
 		 (setq address-list nil))
+		((eq task 'ignore)
+		 (when (cadr address)
+		   (cl-pushnew (downcase (cadr address))
+			       ebdb-permanently-ignored-mails :test #'equal))
+		 (unless ebdb-permanent-ignores-file
+		   (message "Mail will be ignored for this session only")
+		   (sit-for 2)))
+		((eq task 'ignore-domain)
+		 (when (cadr address)
+		   (cl-pushnew (downcase
+				(concat "@"
+					(nth 1 (split-string
+						(cadr address) "@" t))))
+			       ebdb-permanently-ignored-mails :test #'equal)
+		   (unless ebdb-permanent-ignores-file
+		     (message "Mail will be ignored for this session only")
+		     (sit-for 2))))
 		((not (eq task 'next))
 		 (dolist (hit (delq nil (nreverse hits)))
 		   (cl-pushnew hit records :test #'equal)
@@ -758,7 +853,6 @@ Usually this function is called by the wrapper `ebdb-mua-auto-update'."
 
     records))
 
-;;; This whole thing could probably be replaced by `map-y-or-n-p'
 (defun ebdb-query-create ()
   "Interactive query used by `ebdb-update-records'.
 Return t if the record should be created or nil otherwise.
@@ -768,7 +862,7 @@ Honor previous answers such as `!'."
     ;; `ebdb-offer-to-create' holds a character, i.e., a number.
     ;; -- Right now, we only remember "!".
     (when (not (integerp task))
-      (let ((prompt (format "%s is not in EBDB; add? (y,!,n,s,q,?) "
+      (let ((prompt (format "%s is not in EBDB; add? (y,!,n,i,I,s,q,?) "
                             (or (nth 0 ebdb-update-records-address)
                                 (nth 1 ebdb-update-records-address))))
             event)
@@ -789,6 +883,10 @@ Honor previous answers such as `!'."
           ((or (eq task ?q)
                (eq task ?\a)) ; ?\a = C-g
            (throw 'done 'quit))
+	  ((eq task ?i)
+	   (throw 'done 'ignore))
+	  ((eq task ?I)
+	   (throw 'done 'ignore-domain))
           ((eq task ?s)
            (setq ebdb-update-records-p 'existing)
            (throw 'done 'next))
@@ -807,7 +905,9 @@ Honor previous answers such as `!'."
 Type ?  for this help.
 Type y  to add the current record.
 Type !  to add all remaining records.
-Type n  to skip the current record. (You might also type space)
+Type n  to skip the current record. (You can also type space)
+Type i  to permanently ignore this mail address
+Type I  to permanently ignore this mail domain
 Type s  to switch from annotate to search mode.
 Type q  to quit updating records.  No more search or annotation is done.")
                    (set-buffer-modified-p nil)
@@ -830,206 +930,218 @@ a new record is created for ADDRESS.  UPDATE-P may take the values:
  a function     This functions will be called with no arguments.
                   It should return one of the above values.
 Return the records matching ADDRESS or nil."
-  (let* ((mail (nth 1 address))		; possibly nil
-         (name (unless (equal mail (car address))
-                 (car address)))
-	 (record-class (if (eql (nth 3 address) 'organization)
-			   'ebdb-record-organization
-			 ebdb-default-record-class))
-         (records (ebdb-message-search name mail))
-         created-p new-records)
-    (if (and (not records) (functionp update-p))
-        (setq update-p (funcall update-p)))
-    (cond ((eq t update-p) (setq update-p 'create))
-          ((not update-p) (setq update-p 'update)))
+  (pcase-let ((`(,name ,mail ,_header ,header-type ,_mode) address))
+    (let ((record-class (if (eql header-type 'organization)
+			    'ebdb-record-organization
+			  ebdb-default-record-class))
+          (records (ebdb-message-search
+		    name
+		    ;; If `mail-extract-address-components' can't find
+		    ;; a mail address it returns two identical strings
+		    ;; (the name), I don't know why.  But when it
+		    ;; does, EBDB assumes the string is a valid mail
+		    ;; address and tries to find/add it.
+		    (setq mail
+			  (unless (string= mail name)
+			    mail))))
+          created-p new-records)
+      (if (and (not records) (functionp update-p))
+          (setq update-p (funcall update-p)))
+      (cond ((eq t update-p) (setq update-p 'create))
+            ((not update-p) (setq update-p 'update)))
 
-    ;; Create a new record if nothing else fits.
-    ;; In this way, we can fill the slots of the new record with
-    ;; the same code that updates the slots of existing records.
-    (unless (or records
-                (eq update-p 'update)
-                (not (or name mail)))
-      ;; If there is no name, try to use the mail address as name
-      (if (and ebdb-message-mail-as-name mail
-               (or (null name)
-                   (string= "" name)))
+      ;; Create a new record if nothing else fits.
+      ;; In this way, we can fill the slots of the new record with
+      ;; the same code that updates the slots of existing records.
+      (unless (or records
+                  (eq update-p 'update)
+                  (not (or name mail)))
+	;; If there is no name, try to use the mail address as name
+	(when (and ebdb-message-mail-as-name mail
+		   (or (null name)
+                       (string= "" name)))
           (setq name (funcall ebdb-message-clean-name-function mail)))
-      (if (or (eq update-p 'create)
-              (and (eq update-p 'query)
-                   (y-or-n-p (format "%s is not in the EBDB.  Add? "
-                                     (or name mail)))))
-          (setq records (list (ebdb-db-add-record
-			       (car ebdb-db-list)
-			       (make-instance
-				record-class)))
-                created-p t)))
+	(when (or (eq update-p 'create)
+		  (and (eq update-p 'query)
+                       (y-or-n-p (format "%s is not in the EBDB.  Add? "
+					 (or name mail)))))
+          (setq records (list (make-instance record-class))
+                created-p t)
+	  (ebdb-db-add-record (ebdb-prompt-for-db nil t) (car records))
+	  (run-hook-with-args 'ebdb-create-hook (car records))
+	  (run-hook-with-args 'ebdb-change-hook (car records))))
 
-    (dolist (record records)
-      (let* ((old-name (ebdb-record-name record))
-             (mail mail) ;; possibly changed below
-             (created-p created-p)
-             (update-p update-p)
-             change-p add-mails add-name ignore-redundant)
+      (dolist (record records)
+	(let* ((old-name (ebdb-record-name-string record))
+               (mail mail) ;; possibly changed below
+               (created-p created-p)
+               (update-p update-p)
+               change-p add-mails add-name ignore-redundant)
 
-        ;; Analyze the name part of the record.
-        (cond (created-p		; new record
-               (ebdb-record-change-name record name))
+          ;; Analyze the name part of the record.
+          (cond (created-p		; new record
+		 (ebdb-record-change-name record name))
 
-              ((or (not name)
-                   ;; The following tests can differ for more complicated names
-                   (ebdb-string= name old-name)
-                   (ebdb-record-search record 'ebdb-field-name name)))
+		((or (not name)
+                     ;; The following tests can differ for more complicated names
+                     (ebdb-string= name old-name)
+                     (ebdb-record-search record 'ebdb-field-name name)))
 
-              ((null (setq add-name (ebdb-add-job ebdb-add-name record name)))) ; do nothing
+		((null (setq add-name (ebdb-add-job ebdb-add-name record name)))) ; do nothing
 
 
-              ((numberp add-name)
-               (unless ebdb-silent
-                 (message "name mismatch: \"%s\" changed to \"%s\""
-                          old-name name)
-                 (sit-for add-name)))
+		((numberp add-name)
+		 (unless ebdb-silent
+                   (message "name mismatch: \"%s\" changed to \"%s\""
+                            old-name name)
+                   (sit-for add-name)))
 
-              ((ebdb-eval-spec add-name
-                               (if old-name
-                                   (format "Change name \"%s\" to \"%s\"? "
-                                           old-name name)
-                                 (format "Assign name \"%s\" to address \"%s\"? "
-                                         name (ebdb-record-one-mail record))))
-               ;; Keep old-name as AKA?
-               (when (and old-name
-			  ;; Leaky abstraction
-			  (object-of-class-p record 'ebdb-record-person)
-                          (not (member-ignore-case old-name (ebdb-record-alt-names record))))
-                 (if (ebdb-eval-spec (ebdb-add-job ebdb-add-aka record old-name)
-                                     (format "Keep name \"%s\" as an AKA? " old-name))
-                     (ebdb-record-insert-field
-                      record (slot-value record 'name) 'aka)))
-               (ebdb-record-change-name record name)
-               (setq change-p 'name))
+		((ebdb-eval-spec add-name
+				 (if old-name
+                                     (format "Change name \"%s\" to \"%s\"? "
+                                             old-name name)
+                                   (format "Assign name \"%s\" to address \"%s\"? "
+                                           name (ebdb-record-one-mail record))))
+		 ;; Keep old-name as AKA?
+		 (when (and old-name
+			    ;; Leaky abstraction
+			    (object-of-class-p record 'ebdb-record-person)
+                            (not (member-ignore-case old-name (ebdb-record-alt-names record))))
+                   (if (ebdb-eval-spec (ebdb-add-job ebdb-add-aka record old-name)
+                                       (format "Keep name \"%s\" as an AKA? " old-name))
+                       (ebdb-record-insert-field
+			record (slot-value record 'name) 'aka)))
+		 (ebdb-record-change-name record name)
+		 (setq change-p 'name))
 
-              ;; make new name an AKA?
-              ((and old-name
-		    (object-of-class-p record 'ebdb-record-person)
-                    (not (member-ignore-case name (ebdb-record-alt-names record)))
-                    (ebdb-eval-spec (ebdb-add-job ebdb-add-aka record name)
-                                    (format "Make \"%s\" an alternate for \"%s\"? "
-                                            name old-name)))
-               (ebdb-record-insert-field
-                record (ebdb-parse 'ebdb-field-name name) 'aka)
-               (setq change-p 'name)))
+		;; make new name an AKA?
+		((and old-name
+		      (object-of-class-p record 'ebdb-record-person)
+                      (not (member-ignore-case name (ebdb-record-alt-names record)))
+                      (ebdb-eval-spec (ebdb-add-job ebdb-add-aka record name)
+                                      (format "Make \"%s\" an alternate for \"%s\"? "
+                                              name old-name)))
+		 (ebdb-record-insert-field
+                  record (ebdb-parse 'ebdb-field-name name) 'aka)
+		 (setq change-p 'name)))
 
-        ;; Is MAIL redundant compared with the mail addresses
-        ;; that are already known for RECORD?
-        (if (and mail
-                 (setq ignore-redundant
-                       (ebdb-add-job ebdb-ignore-redundant-mails record mail)))
-            (let ((mails (ebdb-record-mail-canon record))
-                  (case-fold-search t) redundant ml re)
-              (while (setq ml (pop mails))
-                (if (and (setq re (ebdb-mail-redundant-re ml))
-                         (string-match re mail))
-                    (setq redundant ml mails nil)))
-              (if redundant
-                  (cond ((numberp ignore-redundant)
-                         (unless ebdb-silent
-                           (message "%s: redundant mail `%s'"
-                                    (ebdb-string record) mail)
-                           (sit-for ignore-redundant)))
-                        ((or (eq t ignore-redundant)
-                             ebdb-silent
-                             (y-or-n-p (format "Ignore redundant mail %s? " mail)))
-                         (setq mail redundant))))))
-	(setq mail (make-instance ebdb-default-mail-class :mail mail))
-        ;; Analyze the mail part of the new records
-        (cond ((or (not mail) (equal (ebdb-string mail) "???")
-                   (member-ignore-case (ebdb-string mail) (ebdb-record-mail-canon record)))) ; do nothing
+          ;; Is MAIL redundant compared with the mail addresses
+          ;; that are already known for RECORD?
+          (if (and mail
+                   (setq ignore-redundant
+			 (ebdb-add-job ebdb-ignore-redundant-mails record mail)))
+              (let ((mails (ebdb-record-mail-canon record))
+                    (case-fold-search t) redundant ml re)
+		(while (setq ml (pop mails))
+                  (if (and (setq re (ebdb-mail-redundant-re ml))
+                           (string-match re mail))
+                      (setq redundant ml mails nil)))
+		(if redundant
+                    (cond ((numberp ignore-redundant)
+                           (unless ebdb-silent
+                             (message "%s: redundant mail `%s'"
+                                      (ebdb-string record) mail)
+                             (sit-for ignore-redundant)))
+                          ((or (eq t ignore-redundant)
+                               ebdb-silent
+                               (y-or-n-p (format "Ignore redundant mail %s? " mail)))
+                           (setq mail redundant))))))
+	  (setq mail (ignore-errors (ebdb-parse ebdb-default-mail-class mail)))
+          ;; Analyze the mail part of the new records
+          (cond ((or (not mail) (equal (ebdb-string mail) "???")
+                     (member-ignore-case (ebdb-string mail) (ebdb-record-mail-canon record)))) ; do nothing
 
-              (created-p		; new record
-               (ebdb-record-insert-field record mail 'mail))
+		(created-p		; new record
+		 (ebdb-record-insert-field record mail 'mail))
 
-              ((not (setq add-mails (ebdb-add-job ebdb-add-mails record mail)))) ; do nothing
+		((not (setq add-mails (ebdb-add-job ebdb-add-mails record mail)))) ; do nothing
 
-              ((numberp add-mails)
-               (unless ebdb-silent
-                 (message "%s: new address `%s'"
-                          (ebdb-string record) (ebdb-string mail))
-                 (sit-for add-mails)))
+		((numberp add-mails)
+		 (unless ebdb-silent
+                   (message "%s: new address `%s'"
+                            (ebdb-string record) (ebdb-string mail))
+                   (sit-for add-mails)))
 
-              ((or (eq add-mails t)	; add it automatically
-                   ebdb-silent
-                   (y-or-n-p (format "Add address \"%s\" to %s? " (ebdb-string mail)
-                                     (ebdb-string record)))
-                   (and (or (and (functionp update-p)
-                                 (progn (setq update-p (funcall update-p)) nil))
-                            (memq update-p '(t create))
-                            (and (eq update-p 'query)
-                                 (y-or-n-p
-                                  (format "Create a new record for %s? "
-                                          (ebdb-string record)))))
-                        (progn
-                          (setq record (make-instance ebdb-default-record-class))
-			  (ebdb-db-add-record (car ebdb-db-list) record)
-                          (ebdb-record-change-name record name)
-                          (setq created-p t))))
+		((or (eq add-mails t)	; add it automatically
+                     ebdb-silent
+                     (y-or-n-p (format "Add address \"%s\" to %s? " (ebdb-string mail)
+                                       (ebdb-string record)))
+                     (and (or (and (functionp update-p)
+                                   (progn (setq update-p (funcall update-p)) nil))
+                              (memq update-p '(t create))
+                              (and (eq update-p 'query)
+                                   (y-or-n-p
+                                    (format "Create a new record for %s? "
+                                            (ebdb-string record)))))
+                          (progn
+                            (setq record (make-instance ebdb-default-record-class))
+			    (ebdb-db-add-record (ebdb-prompt-for-db nil t) record)
+			    (run-hook-with-args 'ebdb-create-hook record)
+			    (run-hook-with-args 'ebdb-change-hook record)
 
-               (let ((mails (ebdb-record-mail record)))
-                 (if ignore-redundant
-                     ;; Does the new address MAIL make an old address redundant?
-                     (let ((mail-re (ebdb-mail-redundant-re (ebdb-string mail)))
-                           (case-fold-search t) okay redundant)
-                       (dolist (ml mails)
-                         (if (string-match mail-re (ebdb-string ml)) ; redundant mail address
-                             (push ml redundant)
-                           (push ml okay)))
-                       (let ((form (format "redundant mail%s %s"
-                                           (if (< 1 (length redundant)) "s" "")
-                                           (ebdb-concat 'mail (nreverse redundant))))
-                             (name (ebdb-record-name record)))
-                         (if redundant
-                             (cond ((numberp ignore-redundant)
-                                    (unless ebdb-silent
-                                      (message "%s: %s" name form)
-                                      (sit-for ignore-redundant)))
-                                   ((or (eq t ignore-redundant)
-                                        ebdb-silent
-                                        (y-or-n-p (format "Delete %s? " form)))
-                                    (if (eq t ignore-redundant)
-                                        (message "%s: deleting %s" name form))
-                                    (setq mails okay)))))))
+                            (ebdb-record-change-name record name)
+                            (setq created-p t))))
 
-                 ;; then modify RECORD
+		 (let ((mails (ebdb-record-mail record)))
+                   (if ignore-redundant
+                       ;; Does the new address MAIL make an old address redundant?
+                       (let ((mail-re (ebdb-mail-redundant-re (ebdb-string mail)))
+                             (case-fold-search t) okay redundant)
+			 (dolist (ml mails)
+                           (if (string-match mail-re (ebdb-string ml)) ; redundant mail address
+                               (push ml redundant)
+                             (push ml okay)))
+			 (let ((form (format "redundant mail%s %s"
+                                             (if (< 1 (length redundant)) "s" "")
+                                             (ebdb-concat 'mail (nreverse redundant))))
+                               (name (ebdb-record-name-string record)))
+                           (if redundant
+                               (cond ((numberp ignore-redundant)
+                                      (unless ebdb-silent
+					(message "%s: %s" name form)
+					(sit-for ignore-redundant)))
+                                     ((or (eq t ignore-redundant)
+                                          ebdb-silent
+                                          (y-or-n-p (format "Delete %s? " form)))
+                                      (if (eq t ignore-redundant)
+                                          (message "%s: deleting %s" name form))
+                                      (setq mails okay)))))))
 
-		 ;; TODO: Reinstate the question about making this primary.
-                 (ebdb-record-insert-field record mail 'mail)
-                 (unless change-p (setq change-p t)))))
+                   ;; then modify RECORD
 
-        (cond (created-p
-               (unless ebdb-silent
-                 (if (ebdb-record-name record)
-                     (message "created %s's record with address \"%s\""
-                              (ebdb-string record)
-			      (ebdb-string mail))
-                   (message "created record with naked address \"%s\""
-			    (ebdb-string mail))))
-               (ebdb-init-record record))
+		   ;; TODO: Reinstate the question about making this primary.
+                   (ebdb-record-insert-field record mail 'mail)
+                   (unless change-p (setq change-p t)))))
 
-              (change-p
-               (unless ebdb-silent
-                 (cond ((eq change-p 'name)
-                        (message "noticed \"%s\"" (ebdb-string record)))
-                       ((ebdb-record-name record)
-                        (message "noticed %s's address \"%s\""
-                                 (ebdb-string record)
-				 (ebdb-string mail)))
-                       (t
-                        (message "noticed naked address \"%s\""
-				 (ebdb-string mail)))))))
+          (cond (created-p
+		 (unless ebdb-silent
+                   (if (ebdb-record-name-string record)
+                       (message "created %s's record with address \"%s\""
+				(ebdb-string record)
+				(ebdb-string mail))
+                     (message "created record with naked address \"%s\""
+			      (ebdb-string mail))))
+		 (ebdb-init-record record)
+		 (run-hook-with-args 'ebdb-after-change-hook record))
 
-        (run-hook-with-args 'ebdb-notice-mail-hook record)
+		(change-p
+		 (unless ebdb-silent
+                   (cond ((eq change-p 'name)
+                          (message "noticed \"%s\"" (ebdb-string record)))
+			 ((ebdb-record-name-string record)
+                          (message "noticed %s's address \"%s\""
+                                   (ebdb-string record)
+				   (ebdb-string mail)))
+			 (t
+                          (message "noticed naked address \"%s\""
+				   (ebdb-string mail)))))
+		 (run-hook-with-args 'ebdb-after-change-hook record)))
 
-        (push record new-records)))
+          (run-hook-with-args 'ebdb-notice-mail-hook record)
 
-    (nreverse new-records)))
+          (push record new-records)))
+
+      (nreverse new-records))))
 
 (cl-defmethod ebdb-mua-prepare-article ()
   "Do whatever preparations are necessary to work on records
@@ -1154,19 +1266,24 @@ where it was in the MUA, rather than quitting the EBDB buffer."
     (set-transient-map
      key-m
      (lambda ()
-       ;; Keep the transient map active until the user hits "q".
-       (null
-	(equal (this-command-keys-vector)
-	       [?q]))))))
+       ;; Keep the transient map active until the user hits "q", but
+       ;; not during minibuffer input.
+       (or (minibufferp)
+	   (null
+	    (equal (this-command-keys-vector)
+		[?q])))))))
 
 ;;;###autoload
-(defun ebdb-mua-toggle-records-format ()
-  "Toggle format of all records without leaving MUA."
-  (interactive)
+(defun ebdb-mua-toggle-records-format (&optional arg)
+  "Toggle format of all records without leaving MUA.
+See the docstring of `ebdb-toglge-records-format' for use of the
+prefix arg ARG."
+  (interactive "p")
   (let ((buf (get-buffer (ebdb-make-buffer-name))))
     (when buf
-     (with-current-buffer buf
-       (ebdb-toggle-records-format ebdb-records)))))
+      (with-current-buffer buf
+	(when ebdb-records
+	  (ebdb-toggle-records-format ebdb-records arg))))))
 
 ;;;###autoload
 (defun ebdb-mua-edit-sender-notes ()
@@ -1264,7 +1381,7 @@ buffer."
 ;; Functions for noninteractive use in MUA hooks
 
 ;;;###autoload
-(defun ebdb-mua-auto-update (&optional header-class update-p)
+(defun ebdb-mua-auto-update (&optional update-p header-class)
   "Update EBDB automatically based on incoming and outgoing messages.
 This looks into the headers of a message according to
 HEADER-CLASS.  Then for the mail addresses found the
@@ -1273,10 +1390,11 @@ whether only existing EBDB records are taken or whether also new
 records are created for these mail addresses.  Return matching
 records.
 
-HEADER-CLASS is defined in `ebdb-message-headers'.  If it is nil,
-use all classes in `ebdb-message-headers'.  UPDATE-P may take the
-same values as `ebdb-mua-auto-update-p'.  If UPDATE-P is nil, use
-`ebdb-mua-auto-update-p' (which see).
+UPDATE-P may take the same values as `ebdb-mua-auto-update-p' or
+any of the MUA-specific equivalents.  If UPDATE-P is nil, use
+`ebdb-mua-auto-update-p' (which see).  HEADER-CLASS is defined in
+`ebdb-message-headers'.  If it is nil, use all classes in
+`ebdb-message-headers'.
 
 If `ebdb-mua-pop-up' is non-nil, EBDB pops up the *EBDB* buffer
 along with the MUA window(s), displaying the matching records."
