@@ -1,6 +1,6 @@
 ;;; ebdb-org.el --- Org mode integration for EBDB    -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2016-2018  Free Software Foundation, Inc.
+;; Copyright (C) 2016-2020  Free Software Foundation, Inc.
 
 ;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>
 ;; Keywords:
@@ -53,9 +53,14 @@
 ;;; Code:
 
 (require 'ebdb-com)
+(require 'ebdb-format)
 (require 'org)
 (require 'org-agenda)
 
+(defgroup ebdb-org nil
+  "Custom group for EBDB Org options."
+  :group 'ebdb)
+
 (if (fboundp 'org-link-set-parameters)
     (org-link-set-parameters "ebdb"
 			     :follow 'ebdb-org-open
@@ -66,8 +71,8 @@
 			     :store 'ebdb-org-store-link
 			     :export 'ebdb-org-export)
   (with-no-warnings ;; I know it's obsolete.
-   (org-add-link-type "ebdb" #'ebdb-org-open #'ebdb-org-export)
-   (add-hook 'org-store-link-functions 'ebdb-org-store-link)))
+    (org-add-link-type "ebdb" #'ebdb-org-open #'ebdb-org-export)
+    (add-hook 'org-store-link-functions 'ebdb-org-store-link)))
 
 ;; TODO: Put a custom keymap on the links (or else expand
 ;; `ebdb-org-open') so that users can choose what to do with the
@@ -78,10 +83,14 @@
   (when (eq major-mode 'ebdb-mode)
     (let* ((rec (ebdb-current-record))
 	   (uuid (ebdb-record-uuid rec))
-	   (name (ebdb-record-name rec))
+	   (name (ebdb-record-name-string rec))
 	   (link (format "ebdb:uuid/%s" uuid)))
-      (org-store-link-props :type "ebdb" :name name
-			    :link link :description name)
+      (with-no-warnings
+	(funcall (if (fboundp 'org-link-store-props)
+		     #'org-link-store-props
+		   #'org-store-link-props)
+		 :type "ebdb" :name name
+		 :link link :description name))
       link)))
 
 (defun ebdb-org-open (link)
@@ -99,9 +108,13 @@
     (`("phone" ,key) (ebdb-search (ebdb-records) `((ebdb-field-phone ,key))))
     (`("address" ,key) (ebdb-search (ebdb-records) `((ebdb-field-address ,key))))
     (`("notes" ,key) (ebdb-search (ebdb-records) `((ebdb-field-notes ,key))))
-    (`("tags" ,key) (ebdb-search (ebdb-records) `((ebdb-org-field-tags ,key))))
-    (`(,(and field (guard (child-of-class-p (intern-soft field) 'ebdb-field))) ,key)
-     (ebdb-search (ebdb-records) `((,(intern-soft field) ,key))))
+    (`("tags" ,key) (ebdb-search (ebdb-records) `((ebdb-field-tags ,key))))
+    (`(,(and field
+	     (let field-sym (intern-soft field))
+	     (and field-sym
+		  (guard (child-of-class-p field-sym 'ebdb-field))))
+       ,key)
+     (ebdb-search (ebdb-records) `((,field-sym ,key))))
     (`(,other _) (error "Unknown field search prefix: %s" other))))
 
 (defun ebdb-org-export (path desc format)
@@ -117,26 +130,35 @@ italicized, in all other cases it is left unchanged."
     (format "<text:span text:style-name=\"Emphasis\">%s</text:span>" desc))
    (t desc)))
 
+;; It was a mistake to make this a separate field class -- this
+;; library should have just provided a new `ebdb-read' method for the
+;; underlying `ebdb-field-tags' class.  I'm overriding `make-instance'
+;; to redirect to `ebdb-field-tags', and will leave this override in
+;; place for a year or so, then remove this class altogether some time
+;; around Feb 2021.
 ;;;###autoload
 (defclass ebdb-org-field-tags (ebdb-field-tags)
   nil
   :human-readable "org tags")
 
-;; Use fast lookups on org-tags, too.
-(cl-pushnew (cons 'ebdb-org-field-tags
-		  (lambda (str rec)
-		    (ebdb-record-search rec 'ebdb-org-field-tags str)))
-	    ebdb-hash-extra-predicates)
+(cl-defmethod make-instance :around ((_cls (subclass ebdb-org-field-tags))
+				     &rest slots)
+  "Return an instance of `ebdb-field-tags' instead."
+  (apply #'cl-call-next-method 'ebdb-field-tags slots))
 
-(cl-defmethod ebdb-read ((field (subclass ebdb-org-field-tags)) &optional slots obj)
+(cl-defmethod ebdb-read ((field (subclass ebdb-field-tags)) &optional slots obj)
   (let* ((crm-separator (cadr (assq 'ebdb-field-tags ebdb-separator-alist)))
 	 (val (completing-read-multiple
-	       "Tags: "
-	       (append (org-global-tags-completion-table)
-		       (when ebdb-tags
-			 (mapcar #'list ebdb-tags)))
+	       (format "Tags (separate with \"%s\"): " crm-separator)
+	       (org--tag-add-to-alist
+		(org--tag-add-to-alist
+		 (org--tag-add-to-alist
+		  (org-global-tags-completion-table)
+		  org-tag-alist)
+		 org-tag-persistent-alist)
+		ebdb-tags)
 	       nil nil
-	       (when obj (ebdb-string obj)))))
+	       (when obj (ebdb-string obj)) 'org-tags-history)))
     (cl-call-next-method field (plist-put slots :tags val))))
 
 ;;;###autoload
@@ -161,5 +183,53 @@ To do this automatically for every search, add this function to
   "Use a separate EBDB buffer for Org-related contacts."
   (format "*%s-Org*" ebdb-buffer-name))
 
+;;; Formatters
+
+(defclass ebdb-org-formatter (ebdb-formatter)
+  ((post-format-function :initform #'org-mode))
+  :abstract t
+  :documentation "Formatter responsible for Org-specific field
+  formatting.")
+
+(cl-defmethod ebdb-fmt-field ((_fmt ebdb-org-formatter)
+			      (_field ebdb-field-mail)
+			      _style
+			      (_rec ebdb-record))
+  (concat "mailto:" (cl-call-next-method)))
+
+(defun ebdb-org-table-post-format ()
+  "Align the formatted Org table."
+  (org-mode)
+  (goto-char (point-min))
+  (forward-char)
+  (org-table-align))
+
+(defclass ebdb-org-formatter-tabular (ebdb-formatter-tabular
+				      ebdb-org-formatter)
+  ((record-separator :initform "\n")
+   (field-separator :initform " | ")
+   (post-format-function :initform #'ebdb-org-table-post-format)))
+
+(cl-defmethod ebdb-fmt-header :around ((_fmt ebdb-org-formatter-tabular)
+				       _records)
+  (concat "| "
+	  (cl-call-next-method)
+	  " |\n"
+	  "|---|\n"))
+
+(cl-defmethod ebdb-fmt-compose-fields :around ((_fmt ebdb-org-formatter-tabular)
+					       (_rec ebdb-record)
+					       &optional _field-list _depth)
+  (concat "| "
+	  (cl-call-next-method)
+	  " |"))
+
+(defcustom ebdb-org-default-tabular-formatter
+  (make-instance 'ebdb-org-formatter-tabular
+		 :label "org table"
+		 :fields '(mail-primary))
+  "Default Org table formatter."
+  :type 'ebdb-formatter-tabular)
+
 (provide 'ebdb-org)
 ;;; ebdb-org.el ends here
