@@ -1,6 +1,6 @@
 ;;; ebdb-pgp.el --- Interaction between EBDB and PGP  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2016-2018  Free Software Foundation, Inc.
+;; Copyright (C) 2016-2020  Free Software Foundation, Inc.
 
 ;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>
 
@@ -131,7 +131,7 @@ See info node `(message)security'."
 
 (cl-defmethod ebdb-read ((class (subclass ebdb-field-pgp)) &optional slots obj)
   (let ((val (intern (ebdb-read-string
-		      "PGP action: " (when obj (slot-value obj 'action))
+		      "PGP action" (when obj (slot-value obj 'action))
 		      ebdb-pgp-ranked-actions t))))
     (cl-call-next-method class (plist-put slots :action val) obj)))
 
